@@ -63,9 +63,32 @@ parse_window_size(const char *json) {
     return size;
 }
 
+bool
+check_host_origin(struct lws *wsi) {
+    int origin_length = lws_hdr_total_length(wsi, WSI_TOKEN_ORIGIN);
+    char buf[origin_length + 1];
+    memset(buf, 0, sizeof(buf));
+    int len = lws_hdr_copy(wsi, buf, sizeof(buf), WSI_TOKEN_ORIGIN);
+    if (len > 0) {
+        const char *prot, *address, *path;
+        int port;
+        if (lws_parse_uri(buf, &prot, &address, &port, &path))
+            return false;
+        sprintf(buf, "%s:%d", address, port);
+        int host_length = lws_hdr_total_length(wsi, WSI_TOKEN_HOST);
+        if (host_length != strlen(buf))
+            return false;
+        char host_buf[host_length + 1];
+        memset(host_buf, 0, sizeof(host_buf));
+        len = lws_hdr_copy(wsi, host_buf, sizeof(host_buf), WSI_TOKEN_HOST);
+        return len > 0 && strcasecmp(buf, host_buf) == 0;
+    }
+    return false;
+}
+
 void
 tty_client_destroy(struct tty_client *client) {
-    if (client->exit)
+    if (client->exit || client->pid <= 0)
         return;
 
     // stop event loop
@@ -82,6 +105,10 @@ tty_client_destroy(struct tty_client *client) {
     lwsl_notice("process exited with code %d, pid: %d\n", status, client->pid);
     close(client->pty);
 
+    // free the buffer
+    if (client->buffer != NULL)
+        t_free(client->buffer);
+
     // remove from clients list
     pthread_mutex_lock(&server->lock);
     LIST_REMOVE(client, list);
@@ -152,15 +179,25 @@ int
 callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
              void *user, void *in, size_t len) {
     struct tty_client *client = (struct tty_client *) user;
-    char *data;
     struct winsize *size;
 
     switch (reason) {
+        case LWS_CALLBACK_FILTER_PROTOCOL_CONNECTION:
+            if (server->once && server->client_count > 0) {
+                lwsl_notice("refuse to serve new client due to the --once option.\n");
+                return -1;
+            }
+            if (server->check_origin && !check_host_origin(wsi)) {
+                lwsl_notice("refuse to serve new client from different origin due to the --check-origin option.\n");
+                return -1;
+            }
+            break;
         case LWS_CALLBACK_ESTABLISHED:
             client->exit = false;
             client->initialized = false;
             client->authenticated = false;
             client->wsi = wsi;
+            client->buffer = NULL;
             lws_get_peer_addresses(wsi, lws_get_socket_fd(wsi),
                                    client->hostname, sizeof(client->hostname),
                                    client->address, sizeof(client->address));
@@ -222,8 +259,18 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
             break;
 
         case LWS_CALLBACK_RECEIVE:
-            data = (char *) in;
-            char command = data[0];
+            if (client->buffer == NULL) {
+                client->buffer = t_malloc(len + 1);
+                client->len = len;
+                memcpy(client->buffer, in, len);
+            } else {
+                client->buffer = t_realloc(client->buffer, client->len + len + 1);
+                memcpy(client->buffer + client->len, in, len);
+                client->len += len;
+            }
+            client->buffer[client->len] = '\0';
+
+            const char command = client->buffer[0];
 
             // check auth
             if (server->credential != NULL && !client->authenticated && command != JSON_DATA) {
@@ -231,9 +278,16 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
                 return -1;
             }
 
+            // check if there are more fragmented messages
+            if (lws_remaining_packet_payload(wsi) > 0 || !lws_is_final_fragment(wsi)) {
+                return 0;
+            }
+
             switch (command) {
                 case INPUT:
-                    if (write(client->pty, data + 1, len - 1) < len - 1) {
+                    if (server->readonly)
+                        return 0;
+                    if (write(client->pty, client->buffer + 1, client->len - 1) < client->len - 1) {
                         lwsl_err("write INPUT to pty\n");
                         return -1;
                     }
@@ -248,7 +302,7 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
                     }
                     break;
                 case RESIZE_TERMINAL:
-                    size = parse_window_size(data + 1);
+                    size = parse_window_size(client->buffer + 1);
                     if (size != NULL) {
                         if (ioctl(client->pty, TIOCSWINSZ, size) == -1) {
                             lwsl_err("ioctl TIOCSWINSZ: %d (%s)\n", errno, strerror(errno));
@@ -260,7 +314,7 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
                     if (server->credential == NULL)
                         break;
                     {
-                        json_object *obj = json_tokener_parse(data);
+                        json_object *obj = json_tokener_parse(client->buffer);
                         struct json_object *o = NULL;
                         if (json_object_object_get_ex(obj, "AuthToken", &o)) {
                             const char *token = json_object_get_string(o);
@@ -276,11 +330,22 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
                     lwsl_notice("unknown message type: %c\n", command);
                     break;
             }
+
+            if (client->buffer != NULL) {
+                t_free(client->buffer);
+                client->buffer = NULL;
+            }
             break;
 
         case LWS_CALLBACK_CLOSED:
             tty_client_destroy(client);
             lwsl_notice("client disconnected from %s (%s), total: %d\n", client->hostname, client->address, server->client_count);
+            if (server->once && server->client_count == 0) {
+                lwsl_notice("exiting due to the --once option.\n");
+                force_exit = true;
+                lws_cancel_service(context);
+                exit(0);
+            }
             break;
 
         default:
